#!/usr/bin/env python3

import re
from datetime import datetime

import requests
from lxml import etree
from peewee import (
    BooleanField,
    CharField,
    CompositeKey,
    DateTimeField,
    IntegerField,
    Model,
    MySQLDatabase,
    TextField,
)

from .config import Config

database = MySQLDatabase(None)


class HIDEvent(Model):
    doorName = CharField()
    timestamp = DateTimeField()
    eventType = IntegerField()
    readerAddress = IntegerField()
    cardholderID = IntegerField(null=True)
    commandStatus = BooleanField(null=True)
    forename = TextField(null=True)
    surname = TextField(null=True)
    ioState = BooleanField(null=True)
    newTime = DateTimeField(null=True)
    oldTime = DateTimeField(null=True)
    rawCardNumber = CharField(max_length=8, null=True)

    class Meta:
        primary_key = CompositeKey("doorName", "timestamp", "eventType")
        database = database


def getStrings(door):
    """Parses out the message strings from source."""
    r = requests.get(
        "https://" + door.ip + "/html/en_EN/en_EN.js",
        auth=requests.auth.HTTPDigestAuth(door.username, door.password),
        verify=False,
    )
    regex = re.compile(r'([0-9]+)="([^"]*)')
    strings = [
        regex.search(s)
        for s in r.text.split(";")
        if s.startswith("localeStrings.eventDetails")
    ]
    print({int(g.group(1)): g.group(2) for g in strings})


@database.atomic()
def getMessages(door):
    last_event = (
        HIDEvent.select(HIDEvent.timestamp)
        .where(HIDEvent.doorName == door.name)
        .order_by(HIDEvent.timestamp.desc())
        .first()
    )
    if last_event is not None:
        last_ts = last_event.timestamp
    else:
        last_ts = datetime(2010, 1, 1)

    events = door.get_events(last_ts)

    HIDEvent.insert_many(
        [
            {
                # fill with None, then overwrite with real contents
                **{k: None for k in HIDEvent._meta.fields.keys()},
                **event.attrib,
                "doorName": door.name,
            }
            for event in events
        ]
    ).on_conflict_ignore().execute()
    return events


def dups(events):
    timestamps = [e.attrib["timestamp"] for e in events]
    dups = set([x for x in timestamps if timestamps.count(x) > 1])
    for e in events:
        if e.attrib["timestamp"] in dups:
            etree.dump(e)


def main():
    config = Config()
    database.init(
        **config.HID_DB,
        **{
            "charset": "utf8",
            "sql_mode": "PIPES_AS_CONCAT",
            "use_unicode": True,
        },
    )
    HIDEvent.create_table()
    for door in config.doors.values():
        getMessages(door)


if __name__ == "__main__":
    main()
