#!/usr/bin/env python3
import asyncio
import random
import re
import string

from udm_rest_client.udm import UDM
from udm_rest_client.exceptions import NoObject, UdmError

from .config import Config

USER_BASE = "cn=users,dc=sawtooth,dc=claremontmakerspace,dc=org"
GROUP_BASE = "cn=groups,dc=sawtooth,dc=claremontmakerspace,dc=org"
GROUPS_REGEX = "|".join(
    ["Certified: .*", "Access .*\?", "CMS .*", "Volunteer: .*", "Database .*"]
)
RAND_PW_LEN = 20


# From an API error message:
#   A group name must start and end with a letter, number or underscore.
#   In between additionally spaces, dashes and dots are allowed.
def sanitize_group_name(name):
    sanitized_body = re.sub(r"[^0-9A-Za-z_ -.]", ".", name)
    sanitized_start_end = re.sub("^[^0-9A-Za-z_]|[^0-9A-Za-z_]$", "_", sanitized_body)

    return "MW_" + sanitized_start_end


# From an API error message: "Username must only contain numbers, letters and dots!"
def sanitize_user_name(name):
    return re.sub(r"[^0-9a-z.]", ".", name.lower()).strip(".")


async def make_groups(group_mod, members):
    existing_group_names = [g.props.name async for g in group_mod.search()]

    groups = [
        sanitize_group_name(group_name)
        for group_name in members[0].keys()
        if re.match(GROUPS_REGEX, group_name) is not None
    ]
    for group_name in groups:
        if group_name not in existing_group_names:
            group = await group_mod.new()
            group.props.name = group_name
            await group.save()


async def _main():
    config = Config()

    members = config.membershipworks.get_members(
        ["Members", "CMS Staff"], "lvl,phn,eml,lbl,nam,end,_id"
    )

    async with UDM(**config.UCS) as udm:
        user_mod = udm.get("users/user")
        group_mod = udm.get("groups/group")

        await make_groups(group_mod, members)

        for member in members:
            username = sanitize_user_name(member["Account Name"])

            try:  # try to get an existing user to update
                user = await user_mod.get(f"uid={username},{USER_BASE}")
            except NoObject:  # create a new user
                # TODO: search by employeeNumber and rename users when needed
                user = await user_mod.new()

                # set a random password and ensure it is changed at next login
                user.props.password = "".join(
                    random.choice(string.ascii_letters + string.digits)
                    for x in range(0, RAND_PW_LEN)
                )
                user.props.pwdChangeNextLogin = True

            user.props.update(
                {
                    "title": "",  # Title
                    "firstname": member["First Name"],
                    "lastname": member["Last Name"],  # (c)
                    "username": username,  # (cmr)
                    "description": "",  # Description
                    # "password": "", # (c) Password
                    # "mailPrimaryAddress": member["Email"], # Primary e-mail address
                    # "displayName": "", # Display name
                    # "birthday": "", # Birthdate
                    # "jpegPhoto": "", # Picture of the user (JPEG format)
                    "employeeNumber": member["Account ID"],
                    # "employeeType": "", # Employee type
                    "homedrive": "H:",  # Windows home drive
                    "sambahome": "\\\\ucs\\" + username,  # Windows home path
                    "profilepath": "%LOGONSERVER%\\%USERNAME%\\windows-profiles\\default",  # Windows profile directory
                    "disabled": member["Account on Hold"] != "",
                    # "userexpiry": member["Renewal Date"],
                    # "pwdChangeNextLogin": "1", # User has to change password on next login
                    # "sambaLogonHours": "", # Permitted times for Windows logins
                    "e-mail": [member["Email"]],  # ([]) E-mail address
                    "phone": [member["Phone"]],  # Telephone number
                    # "PasswordRecoveryMobile": member["Phone"], # Mobile phone number
                    "PasswordRecoveryEmail": member["Email"],
                }
            )

            new_groups = [
                "cn=" + sanitize_group_name(group) + "," + GROUP_BASE
                for group, value in member.items()
                if re.match(GROUPS_REGEX, group) is not None and value != ""
            ]
            # groups not from this script
            other_old_groups = [
                g for g in user.props.groups if not g[3:].startswith("MW_")
            ]
            user.props.groups = other_old_groups + new_groups

            try:
                await user.save()
            except UdmError:
                print("Failed to save user", username)
                print(user.props)
                raise


def main():
    asyncio.run(_main())


if __name__ == "__main__":
    main()
