#!/usr/bin/env python3

from datetime import datetime
import sys

import pyclip

from .config import Config

TIME_FMT = "%l:%M%P"
DATETIME_FMT = "%a %b %e %Y, " + TIME_FMT


def format_datetime_range(start_ts: int, end_ts: int):
    start = datetime.fromtimestamp(start_ts)
    end = datetime.fromtimestamp(end_ts)
    start_str = start.strftime(DATETIME_FMT)
    if start.date() == end.date():
        end_str = end.strftime(TIME_FMT)
    else:
        # TODO: this probably implies multiple instances. Should read
        #       RRULE or similar from the event notes
        end_str = end.strftime(DATETIME_FMT)
    return f"{start_str} &mdash; {end_str}"


def format_event(event_details, truncate: bool):
    try:
        url = (
            "https://claremontmakerspace.org/events/#!event/register/"
            + event_details["url"]
        )
        if "lgo" in event_details:
            img = f"""<img class="alignleft" width="400" src="{event_details['lgo']['l']}">"""
        else:
            img = ""
        # print(json.dumps(event_details))
        out = f"""<h2 style="text-align: center;">
  <a href="{url}">{img}{event_details['ttl']}</a>
</h2>
<div><i>{format_datetime_range(event_details['sdp'], event_details['edp'])}</i></div>
"""
        if not truncate:
            out += f"""
<div>
  {event_details['dtl']}
</div>

<a href="{url}">Register for this class now!</a>"""
        return out
    except KeyError as e:
        print(
            f"Event '{event_details.get('ttl')}' missing required property: '{e.args[0]}'"
        )
        raise


def format_section(title: str, blurb: str, events, truncate: bool):
    # skip empty sections
    if not events:
        return ""

    events_list = "\n<hr />\n\n".join(format_event(event, truncate) for event in events)

    return f"""<h1>{title}</h1>
<h4><i>{blurb}</i></h4>
{events_list}
"""


def generate_post():
    config = Config()
    now = datetime.now()

    membershipworks = config.membershipworks
    events = membershipworks.get_events_list(now)
    if "error" in events:
        print("Error:", events["error"])
        return

    ongoing_events = []
    full_events = []
    upcoming_events = []
    for event in events["evt"]:
        try:
            # ignore hidden events
            if event["cal"] == 0:
                continue
            event_details = membershipworks.get_event_by_eid(event["eid"])

            # registration has already ended
            if (
                "erd" in event_details
                and datetime.fromtimestamp(event_details["erd"]) < now
            ):
                ongoing_events.append(event_details)
            # class is full
            elif event_details["cnt"] >= event_details["cap"]:
                full_events.append(event_details)
            else:
                upcoming_events.append(event_details)
        except KeyError as e:
            print(
                f"Event '{event.get('ttl')}' missing required property: '{e.args[0]}'"
            )
            raise

    # header
    yield """<p><img class="aligncenter size-medium wp-image-2319" src="https://claremontmakerspace.org/wp-content/uploads/2019/03/CMS-Logo-b-y-g-300x168.png" alt="" width="300" height="168" /></a></p>
<p>Greetings Upper Valley Makers:</p>
<p>We have an exciting list of upcoming classes at the Claremont MakerSpace that we think might interest you.</p>
<strong>For most classes and events, CMS MEMBERSHIP IS NOT REQUIRED.</strong> That said, members receive a discount on registration and there are some classes/events that are for members only (this will be clearly noted in the event description).

<strong>Class policies</strong> (liability waiver, withdrawal, cancellation, etc.) can be found <a href="https://claremontmakerspace.org/class-policies/" data-wpel-link="internal">here</a>.

<strong>Instructors:</strong> Interested in teaching a class at CMS? Please fill out our <a href="https://claremontmakerspace.org/cms-class-proposal-form/" data-wpel-link="internal">Class Proposal Form</a>.

<strong>Tours:</strong> Want to see what the Claremont MakerSpace is all about? Tours are by appointment only. <a href="https://tickets.claremontmakerspace.org/open.php" target="_blank" rel="noreferrer noopener external" data-wpel-link="external">Contact Us</a> to schedule your tour where you can learn about all the awesome tools that the CMS offers access to, as well as how membership, classes, and studio spaces work.

<hr />
"""

    yield format_section(
        "Upcoming Events",
        "Events that are currently open for registration.",
        upcoming_events,
        truncate=False,
    )

    yield format_section(
        "<hr />Just Missed",
        "These classes are currently full at time of writing. If you are interested, please check the event's page; spots occasionally open up. Keep an eye on this newsletter to see when these classes are offered again.",
        full_events,
        truncate=True,
    )

    yield format_section(
        "<hr />Ongoing Events",
        "These events are ongoing. Registration is currently closed, but these events may be offered again in the future.",
        ongoing_events,
        truncate=True,
    )

    # footer
    yield """<div style="clear: both;">
  <hr />

  <div>Happy Makin’!</div>
  <div>We are grateful for all of the public support that our 501(c)(3), non-profit organization receives. If you’d like to make a donation,please visit the <a href="https://claremontmakerspace.org/support/"><strong>Support Us page</strong></a> of our website.</div>
</div>
"""


def main():
    result = "\n".join(generate_post())
    print(result)
    pyclip.copy(result)
    print("Copied to clipboard!", file=sys.stderr)


if __name__ == "__main__":
    main()
